package java.util.zip;

import java.io.EOFException;
import java.io.PushbackInputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.io.InputStream;
import java.io.IOException;

public class ZipInputStream extends InflaterInputStream implements ZipConstants
{
    private ZipEntry entry;
    private int flag;
    private CRC32 crc;
    private long remaining;
    private byte[] tmpbuf;
    private static final int STORED = 0;
    private static final int DEFLATED = 8;
    private boolean closed;
    private boolean entryEOF;
    private ZipCoder zc;
    private byte[] b;
    
    private void ensureOpen() throws IOException {
        if (this.closed) {
            throw new IOException("Stream closed");
        }
        return;
    }
    
    public ZipInputStream(InputStream inputStream) {
        this(inputStream, StandardCharsets.UTF_8);
    }
    
    public ZipInputStream(InputStream in, Charset charset) {
        super(new PushbackInputStream(in, 512), new Inflater(true), 512);
        this.crc = new CRC32();
        this.tmpbuf = new byte[512];
        this.closed = false;
        this.entryEOF = false;
        this.b = new byte[256];
        this.usesDefaultInflater = true;
        if (in == null) {
            throw new NullPointerException("in is null");
        }
        if (charset == null) {
            throw new NullPointerException("charset is null");
        }
        this.zc = ZipCoder.get(charset);
        return;
    }
    
    public ZipEntry getNextEntry() throws IOException {
        this.ensureOpen();
        if (this.entry != null) {
            this.closeEntry();
        }
        this.crc.reset();
        this.inf.reset();
        if ((this.entry = this.readLOC()) == null) {
            return null;
        }
        if (this.entry.method == 0) {
            this.remaining = this.entry.size;
        }
        this.entryEOF = false;
        return this.entry;
    }
    
    public void closeEntry() throws IOException {
        this.ensureOpen();
        while (this.read(this.tmpbuf, 0, this.tmpbuf.length) != -1) {}
        this.entryEOF = true;
    }
    
    @Override
    public int available() throws IOException {
        this.ensureOpen();
        if (this.entryEOF) {
            return 0;
        }
        return 1;
    }
    
    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        this.ensureOpen();
        if (off < 0 || len < 0 || off > b.length - len) {
            throw new IndexOutOfBoundsException();
        }
        if (len == 0) {
            return 0;
        }
        if (this.entry == null) {
            return -1;
        }
        switch (this.entry.method) {
            case 8: {
                len = super.read(b, off, len);
                if (len == -1) {
                    this.readEnd(this.entry);
                    this.entryEOF = true;
                    this.entry = null;
                }
                else {
                    this.crc.update(b, off, len);
                }
                return len;
            }
            case 0: {
                if (this.remaining <= 0L) {
                    this.entryEOF = true;
                    this.entry = null;
                    return -1;
                }
                if ((long)len > this.remaining) {
                    len = (int)this.remaining;
                }
                len = this.in.read(b, off, len);
                if (len == -1) {
                    throw new ZipException("unexpected EOF");
                }
                this.crc.update(b, off, len);
                this.remaining -= (long)len;
                if (this.remaining == 0L && this.entry.crc != this.crc.getValue()) {
                    throw new ZipException("invalid entry CRC (expected 0x" + Long.toHexString(this.entry.crc) + " but got 0x" + Long.toHexString(this.crc.getValue()) + ")");
                }
                return len;
            }
            default: {
                throw new ZipException("invalid compression method");
            }
        }
    }
    
    @Override
    public long skip(long a) throws IOException {
        if (a < 0L) {
            throw new IllegalArgumentException("negative skip length");
        }
        this.ensureOpen();
        int n = (int)Math.min(a, 2147483647L);
        int i = 0;
        while (i < n) {
            int length = n - i;
            if (length > this.tmpbuf.length) {
                length = this.tmpbuf.length;
            }
            int read = this.read(this.tmpbuf, 0, length);
            if (read == -1) {
                this.entryEOF = true;
                break;
            }
            i += read;
        }
        return (long)i;
    }
    
    @Override
    public void close() throws IOException {
        if (!this.closed) {
            super.close();
            this.closed = true;
        }
    }
    
    private ZipEntry readLOC() throws IOException {
        try {
            this.readFully(this.tmpbuf, 0, 30);
        }
        catch (EOFException ex) {
            return null;
        }
        if (ZipUtils.get32(this.tmpbuf, 0) != 67324752L) {
            return null;
        }
        this.flag = ZipUtils.get16(this.tmpbuf, 6);
        int i = ZipUtils.get16(this.tmpbuf, 26);
        int length = this.b.length;
        if (i > length) {
            do {
                length *= 2;
            } while (i > length);
            this.b = new byte[length];
        }
        this.readFully(this.b, 0, i);
        ZipEntry zipEntry = this.createZipEntry(((this.flag & 0x800) != 0x0) ? this.zc.toStringUTF8(this.b, i) : this.zc.toString(this.b, i));
        if ((this.flag & 0x1) == 0x1) {
            throw new ZipException("encrypted ZIP entry not supported");
        }
        zipEntry.method = ZipUtils.get16(this.tmpbuf, 8);
        zipEntry.xdostime = ZipUtils.get32(this.tmpbuf, 10);
        if ((this.flag & 0x8) == 0x8) {
            if (zipEntry.method != 8) {
                throw new ZipException("only DEFLATED entries can have EXT descriptor");
            }
        }
        else {
            zipEntry.crc = ZipUtils.get32(this.tmpbuf, 14);
            zipEntry.csize = ZipUtils.get32(this.tmpbuf, 18);
            zipEntry.size = ZipUtils.get32(this.tmpbuf, 22);
        }
        int get16 = ZipUtils.get16(this.tmpbuf, 28);
        if (get16 > 0) {
            byte[] array = new byte[get16];
            this.readFully(array, 0, get16);
            zipEntry.setExtra0(array, zipEntry.csize == 4294967295L || zipEntry.size == 4294967295L);
            return zipEntry;
        }
        return zipEntry;
    }
    
    protected ZipEntry createZipEntry(String s) {
        return new ZipEntry(s);
    }
    
    private void readEnd(ZipEntry zipEntry) throws IOException {
        int remaining = this.inf.getRemaining();
        if (remaining > 0) {
            ((PushbackInputStream)this.in).unread(this.buf, this.len - remaining, remaining);
        }
        if ((this.flag & 0x8) == 0x8) {
            if (this.inf.getBytesWritten() > 4294967295L || this.inf.getBytesRead() > 4294967295L) {
                this.readFully(this.tmpbuf, 0, 24);
                long get33 = ZipUtils.get32(this.tmpbuf, 0);
                if (get33 != 134695760L) {
                    zipEntry.crc = get33;
                    zipEntry.csize = ZipUtils.get64(this.tmpbuf, 4);
                    zipEntry.size = ZipUtils.get64(this.tmpbuf, 12);
                    ((PushbackInputStream)this.in).unread(this.tmpbuf, 19, 4);
                }
                else {
                    zipEntry.crc = ZipUtils.get32(this.tmpbuf, 4);
                    zipEntry.csize = ZipUtils.get64(this.tmpbuf, 8);
                    zipEntry.size = ZipUtils.get64(this.tmpbuf, 16);
                }
            }
            else {
                this.readFully(this.tmpbuf, 0, 16);
                long get32 = ZipUtils.get32(this.tmpbuf, 0);
                if (get32 != 134695760L) {
                    zipEntry.crc = get32;
                    zipEntry.csize = ZipUtils.get32(this.tmpbuf, 4);
                    zipEntry.size = ZipUtils.get32(this.tmpbuf, 8);
                    ((PushbackInputStream)this.in).unread(this.tmpbuf, 11, 4);
                }
                else {
                    zipEntry.crc = ZipUtils.get32(this.tmpbuf, 4);
                    zipEntry.csize = ZipUtils.get32(this.tmpbuf, 8);
                    zipEntry.size = ZipUtils.get32(this.tmpbuf, 12);
                }
            }
        }
        if (zipEntry.size != this.inf.getBytesWritten()) {
            throw new ZipException("invalid entry size (expected " + zipEntry.size + " but got " + this.inf.getBytesWritten() + " bytes)");
        }
        if (zipEntry.csize != this.inf.getBytesRead()) {
            throw new ZipException("invalid entry compressed size (expected " + zipEntry.csize + " but got " + this.inf.getBytesRead() + " bytes)");
        }
        if (zipEntry.crc != this.crc.getValue()) {
            throw new ZipException("invalid entry CRC (expected 0x" + Long.toHexString(zipEntry.crc) + " but got 0x" + Long.toHexString(this.crc.getValue()) + ")");
        }
        return;
    }
    
    private void readFully(byte[] b, int off, int i) throws IOException {
        while (i > 0) {
            int read = this.in.read(b, off, i);
            if (read == -1) {
                throw new EOFException();
            }
            off += read;
            i -= read;
        }
    }
}
